@@ -28,6 +28,15 @@ class TestAdd(unittest.TestCase):
         except Exception as ex:
             self.fail('failed to add bytes: %s' % ex)
 
+    def test_cardinality_estimate_changed_return(self):
+        changed = self.hll.add('asdf')
+        self.assertTrue(changed)
+        changed = self.hll.add('asdf')
+        self.assertFalse(changed)
+        changed = self.hll.add('otherdata')
+        self.assertTrue(changed)
+
+
 class TestCardinalityEstimation(unittest.TestCase):
 
     def setUp(self):
@@ -39,8 +48,8 @@ class TestCardinalityEstimation(unittest.TestCase):
     def test_small_range_correction_not_all_registers_set_to_zero(self):
         self.hll.set_register(0, 1)
         c = self.hll.cardinality()
-        correction= 1.46571806761 <= c and c <= 1.46571806762
-        self.assertTrue(correction)
+        expected = 1.01 <= c and c <= 1.02
+        self.assertTrue(expected)
 
     def test_medium_range_no_correction(self):
         for i in range(32):
@@ -59,7 +68,7 @@ class TestCardinalityEstimation(unittest.TestCase):
         c = hll.cardinality()
         correction = 7916284520 <= c and c <= 7916284521
         self.assertTrue(correction)
- 
+
 class TestHyperLogLogConstructor(unittest.TestCase):
 
     def test_one_is_invalid_size(self):
@@ -75,24 +84,24 @@ class TestHyperLogLogConstructor(unittest.TestCase):
             HyperLogLog(2)
         except Exception:
             self.fail()
-    
+
     def test_maximum_size_is_valid(self):
         try:
             HyperLogLog(16)
         except Exception:
             self.fail()
-    
+
     def test_greater_than_the_maximum_size_is_invalid(self):
         with self.assertRaises(ValueError):
-            HyperLogLog(17)           
-                 
+            HyperLogLog(17)
+
     def test_all_registers_initialized_to_zero(self):
         hll = HyperLogLog(5)
         registers = hll.registers()
         for register in registers:
             self.assertEqual(register, 0)
 
-    def test_k_param_correctly_determines_the_number_of_registers(self):
+    def test_k_param_determines_the_number_of_registers(self):
         hll = HyperLogLog(5)
         self.assertEqual(len(hll.registers()), 32)
         self.assertEqual(hll.size(), 32)
@@ -111,7 +120,7 @@ class TestMerging(unittest.TestCase):
         hll2 = HyperLogLog(5)
         with self.assertRaises(ValueError):
             hll.merge(hll2)
-             
+
     def test_merge(self):
         expected = bytearray(4)
         expected[0] = 1
@@ -129,19 +138,37 @@ class TestMerging(unittest.TestCase):
 class TestPickling(unittest.TestCase):
 
     def setUp(self):
-        self.hll = HyperLogLog(12, seed=123)
-        for i in range(1000000): self.hll.add(str(i))
+        hlls = [HyperLogLog(x, randint(1, 1000)) for x in range(4, 16)]
+        cardinalities = [x**5 for x in range(1, 16)]
 
-    def test_pickle(self):
-        expected = self.hll.cardinality()
-        hll2 = pickle.loads(pickle.dumps(self.hll))
-        self.assertEqual(expected, hll2.cardinality())
+        for hll, n in zip(hlls, cardinalities):
+            for i in range(1, n):
+                hll.add(str(i))
+        self.hlls = hlls
 
-    @unittest.skip('cPickle does not import on travis, import before enabling')
-    def test_cPickle(self):
-        expected = self.hll.cardinality()
-        hll2 = cPickle.loads(cPickle.dumps(self.hll))
-        self.assertEqual(expected, hll2.cardinality())
+    def test_pickled_cardinality(self):
+        for hll in self.hlls:
+            expected = hll.cardinality()
+            hll2 = pickle.loads(pickle.dumps(hll))
+            self.assertEqual(expected, hll2.cardinality())
+
+    def test_pickled_seed(self):
+        for hll in self.hlls:
+            expected = hll.seed()
+            hll2 = pickle.loads(pickle.dumps(hll))
+            self.assertEqual(expected, hll2.seed())
+
+    def test_pickled_registers(self):
+        for hll in self.hlls:
+            expected = hll.registers()
+            hll2 = pickle.loads(pickle.dumps(hll))
+            self.assertEqual(expected, hll2.registers())
+
+    def test_pickled_size(self):
+         for hll in self.hlls:
+            expected = hll.size()
+            hll2 = pickle.loads(pickle.dumps(hll))
+            self.assertEqual(expected, hll2.size())
 
 class TestRegisterFunctions(unittest.TestCase):
 
@@ -173,23 +200,47 @@ class TestRegisterFunctions(unittest.TestCase):
         with self.assertRaises(ValueError):
             self.hll.set_register(0, -1)
 
-    def test_bytesarray_returned_from_registers_contains_correct_values(self):
-        expected = bytearray(32)
-        for i in range(31):
-            expected[i] = randint(0, 16)
-
-        for i in range(31):
+    def test_bytesarray_has_correct_values(self):
+        expected = bytearray(randint(0, 16) for x in range(32))
+        for i in range(32):
             self.hll.set_register(i, expected[i])
 
         registers = self.hll.registers()
-        for i in range(31):
-            self.assertEqual(expected[i], registers[i])
+        self.assertEqual(expected, registers)
 
     def test_registers_returns_bytesarray(self):
         self.assertTrue(type(self.hll.registers()) is bytearray)
 
-    def test_registers_returns_correct_length_bytearray(self):
+    def test_bytesarray_has_correct_length(self):
         self.assertTrue(len(self.hll.registers()) == pow(2, self.k))
 
+    def test_set_registers_with_bytearray(self):
+        expected = bytearray(randint(0, 16) for x in range(32))
+        self.hll.set_registers(expected)
+
+        registers=self.hll.registers()
+        self.assertEqual(expected, registers)
+
+    def test_set_registers_with_bytes(self):
+        expected = bytearray(randint(0, 16) for x in range(32))
+        self.hll.set_registers(bytes(expected))
+
+        registers=self.hll.registers()
+        self.assertEqual(expected, registers)
+
+    @unittest.skipIf(sys.version_info[0] >= 3, 'TypeError raised in Python 3')
+    def test_set_registers_raises_value_error(self):
+        expected = bytearray(range(10, 42))
+
+        with self.assertRaises(ValueError):
+            self.hll.set_registers(bytes(expected))
+
+    @unittest.skipIf(sys.version_info[0] < 3, 'ValueError raised in Python 2')
+    def test_set_registers_raises_type_error(self):
+        expected = bytearray(range(10, 42))
+
+        with self.assertRaises(ValueError):
+            self.hll.set_registers(bytes(expected))
+
 if __name__ == '__main__':
     unittest.main()
